/*
  t2fogskies

  Copyright 2010 Jere Leppnen <jere.leppanen@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#define LOG(level, args...) do { if(par.verbose>level) { printf("%s(%d): ", __FUNCTION__, __LINE__); printf(args); printf("\n"); } } while(0)
#define ERROR(args...) do { printf("%s(%d): ERROR: ", __FUNCTION__, __LINE__); printf(args); printf("\n"); exit(1); } while(0)

#define PROGRAM "t2fogskies"

struct
{
  int verbose, yes;
  char *dir;
} par={0, 0, NULL};

int fileexists(const char *name)
{
  LOG(2, " ");
  FILE *file=fopen(name, "rb");
  if(file) fclose(file);
  LOG(1, "File %s %s.", name, file ? "exists" : "doesn't exist");
  return file!=NULL;
}

void searchdir()
{
  LOG(2, " ");

  int c;
  char *cdir="", *ldir=NULL;
  for(c=0; c<128; c++)
    {
      free(ldir);
      ldir=cdir;
      if(!(cdir=getcwd(NULL, 0))) ERROR("getcwd()");
      LOG(3, "%s", cdir);
      if(!strcmp(ldir, cdir)) break;
      if(fileexists("thief2.exe"))
	{
	  printf("Thief2 directory is %s.\n", cdir);
	  return;
	}
      chdir("..");
    }
  ERROR("Cannot find Thief2 installation directory.");
}

int confirm(const char *msg)
{
  LOG(2, " ");

  if(par.yes) return 1;

  printf(msg);
  fflush(stdout);
  char tmp[128];
  memset(tmp, 0, 128);
  scanf("%s", tmp);
  if(*tmp=='y' || *tmp=='Y') return 1;
  return 0;
}

void cmd_help()
{
  LOG(2, " ");
  printf("\nUsage: %s [options]\n"
	 "\nOptions:\n"
	 "    -d <dir>    search for mis files in directory <dir>\n"
	 "    -h          help\n"
	 "    -v <level>  verbosity level\n"
	 "    -y          don't prompt, assume \"yes\"\n",
	 PROGRAM);
  printf("\n\
This program tweaks some parameters in the Thief 2: The Metal Age\n\
missions Framed, Life of the Party and Masks so that skies and distant\n\
art are visible with ddfix fog.\n\
\n\
Requires ddfix 1.4.3_alpha or later with ddfix.ini settings:\n\
  EnableFogFix=1\n\
  EnableLifeOfThePartyFogFix=1\n\
\n\
The easiest way to use this program is to place it anywhere under your\n\
Thief2 installation directory and then just double-click on\n\
%s You'll be prompted on each patching.\n\
\n\
If the -d option is not used, Thief2 installation directory is\n\
searched by starting at the current working directory and going up the\n\
directory tree level by level until thief2.exe is found.\n\
\n", PROGRAM);
  exit(1);
}

#define BUFSIZE 4096

void patch(const char *name, const char *needle, size_t offset, void *replacement, int rlen)
{
  LOG(2, "name=%s needle=%s offset=%d rlen=%d", name, needle, offset, rlen);

  printf("Processing file %s.\n", name);

  FILE *file=fopen(name, "rb");
  if(!file) 
    {      
      LOG(0, "Cannot open file %s.", name);
      return;
    }
  
  char *buf=malloc(BUFSIZE*2);
  memset(buf, 0, BUFSIZE*2);

  int found=0, block=-1;
  while(!found && !feof(file))
    {
      memcpy(buf, buf+BUFSIZE, BUFSIZE);
      int i=fread(buf+BUFSIZE, 1, BUFSIZE, file);
      memset(buf+BUFSIZE+i, 0, BUFSIZE-i);
      char *p=memchr(buf, *needle, BUFSIZE);
      while(p)
	{
	  if(!memcmp(needle, p, strlen(needle)))
	    {
	      found=1;
	      fclose(file);
	      size_t loc=block*BUFSIZE+(p-buf);
	      LOG(1, "Found %s at 0x%x", needle, loc);
	      if(confirm("\nPatching is not undoable. Information will be lost.\n"
			 "Are you sure you've made backups and want to continue (y/n)? "))
		{
		  file=fopen(name, "r+b");
		  if(!file) ERROR("Cannot open %s for writing.", name);
		  loc+=offset;
		  if(fseek(file, loc, SEEK_SET)) ERROR("Cannot seek to 0x%x in %s.", loc, name);
		  if(fwrite(replacement, rlen, 1, file)!=1) ERROR("Cannot write to %s.", name);
		  printf("%s patched.\n", name);
		}
	      else
		printf("No changes made to %s.\n", name);

	      break;
	    }
	  p=memchr(p+1, *needle, BUFSIZE-(p-buf)-1);
	}
      block++;
    }

  if(!found) printf("Didn't find %s.\n", needle);
  free(buf);
  fclose(file);
}

void mispatch()
{
  LOG(2, " ");

  float value;
  struct {
    float x, y, z;
  } wind_velocity;
  
  // Framed
  // atmosphere radius
  value=50050;
  patch("miss4.mis", "SKYOBJVAR", 32, &value, sizeof(value));
  // cloud height
  value=50;
  patch("miss4.mis", "CLOUDOBJVAR", 296, &value, sizeof(value));
  // cloud tile size
  value=100;
  patch("miss4.mis", "CLOUDOBJVAR", 300, &value, sizeof(value));
  // cloud wind velocity
  wind_velocity.x=.5;
  wind_velocity.y=.5;
  wind_velocity.z=0;
  patch("miss4.mis", "CLOUDOBJVAR", 312, &wind_velocity, sizeof(wind_velocity));

  // Life of the Party
  // distant art alpha
  value=.1;
  patch("miss11.mis", "DISTOBJVAR", 576, &value, sizeof(value));

  // Masks
  // atmosphere radius
  value=50050;
  patch("miss13.mis", "SKYOBJVAR", 32, &value, sizeof(value));
  // cloud height
  value=50;
  patch("miss13.mis", "CLOUDOBJVAR", 296, &value, sizeof(value));
  // cloud tile size
  value=100;
  patch("miss13.mis", "CLOUDOBJVAR", 300, &value, sizeof(value));
  // cloud wind velocity
  wind_velocity.x=1;
  wind_velocity.y=.5;
  wind_velocity.z=0;
  patch("miss13.mis", "CLOUDOBJVAR", 312, &wind_velocity, sizeof(wind_velocity));
  // distant art alpha
  value=.1;
  patch("miss13.mis", "DISTOBJVAR", 576, &value, sizeof(value));

}

int main(int argc, char *argv[])
{
  int c;
  for(c=1; c<argc; c++)
    {
      if(!strncasecmp(argv[c], "-h", 2)) cmd_help();
    }

  opterr=0;
  while(1)
    {
      int o=getopt(argc, argv, "d:v:y");
      if(o<0) break;
      switch(o)
	{
	case 'd':
	  par.dir=strdup(optarg);
	  break;
	case 'v':
	  par.verbose=atoi(optarg);
	  break;
	case 'y':
	  par.yes=1;
	  break;
	case ':': case '?': default:
	  ERROR("Invalid option \'%c\' or missing value. (Try %s -h)", optopt, PROGRAM);
	}
    }

  LOG(1, "%s %s", __FILE__, __DATE__);

  if(par.dir)
    {
      LOG(1, "chdir(%s)", par.dir);
      if(chdir(par.dir)) ERROR("Cannot change to dir %s.", par.dir);
    }
  else searchdir();

  mispatch();

  return 0;
}
